from itertools import imap
from .helpers import traverse, draw_tree, convert_to_snake_case


class Maze(object):

    def __init__(self, graph):
        self.graph = graph

    def _optimal_path(self, is_solution):
        possible_solutions = []

        def on_visit(path):
            if not is_solution(path):
                # current path does not satisfy our constraints, so
                # a possible solution might exist if the current path's
                # last node has children, so we return it to be explored
                return Node.decorate_leaves_with_lineage(path)

            # if we've satisified our reason for traversal
            # then we've hit a potential path!
            possible_solutions.append(path)

        traverse([self.graph.root], on_visit)

        # sort possible solutions on length, since weight is assumed to be one,
        # return smallest.
        possible_solutions.sort(cmp=lambda x, y: len(x) > len(y))
        return possible_solutions[0]

    def route(self, node, include=None):
        """
        Given an directed acyclic graph, ``self.graph``, this method
        will default to finding the shortest path to the desired
        ``node``.

        If ``include`` is passed in, this method will try to find out
        the shortest topologically sorted path that includes a visit
        to all desired nodes in ``include``, eventually, ending up to
        the leaf ``node``.

        """
        include = set(include) if include else set()

        def is_solution(path):
            last_node = path[-1]
            # set of visited nodes in path
            remaining_nodes = set(path[:-1])

            # see if there is any path in current_paths
            # already statifies our requirements
            return last_node == node and remaining_nodes.issuperset(include)

        return self._optimal_path(is_solution)


class Node(object):

    @classmethod
    def decorate_leaves_with_lineage(cls, path):
        """
        Decorates a node's children list with the full absolute path
        to that node.

        :param path: A list containing the last node seen, prepended with its
                     lineage.

        :return: A decorated :ref:`Node` that contains a modified children
                 iterable attribute, where every child node is prepended with
                 its lineage (in this case, the path).
        """
        ln = path[-1]
        decorated = cls(ln.name)
        decorated.children = imap(lambda c: path[:] + [c], ln.children)
        return decorated

    def __init__(self, name=None):
        self.name = name or convert_to_snake_case(self.__class__.__name__)
        self.children = []

    def add_child(self, node):
        self.children.append(node)

    def find(self, child_name):
        if child_name == self.name:
            return self
        for child in self.children:
            if child.name == child_name:
                return child

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'Node(%s)' % self.name

    def draw(self):
        print '\n' + draw_tree(self)


class Graph(object):
    """
    Represents a collection of :ref:`Node`s. Acts as a fascade to operate
    on a set of nodes.

    """
    def __init__(self, root):
        self.root = root
        self._nodes = None

    def draw(self):
        self.root.draw()

    @property
    def nodes(self):
        """
        Returns a unique set of nodes seen after traversing the entire graph.
        """
        if self._nodes:
            return self._nodes

        uniq_nodes = set()

        def on_visit(node):
            uniq_nodes.add(node)
            return node

        traverse(self.root, on_visit)
        self._nodes = uniq_nodes
        return self._nodes



from cStringIO import StringIO
from collections import Iterable
import re


def traverse(start, on_visit):
    """
    Performs a depth-first serach on a tree.

    :param start: The starting node to begin the search
    :param on_visit: A callback function that will be invoked after visiting
                     a potential path.

                     If this method does not return a path, then the search
                     will continue to exhaust the next path in the stack.

                     If a path is returned, it must satisfy a ``children``
                     iterable attribute. This attribute must yield the
                     next paths to push on to the stack.
    :return: None
    """
    paths_to_explore = [start]
    while paths_to_explore:
        path = paths_to_explore.pop(0)
        path = on_visit(path)
        if not path:
            continue

        assert isinstance(getattr(path, 'children'), Iterable), (
            "Path %s must have a iterable attribute 'children'" % path
        )

        for child in path.children:
            paths_to_explore.append(child)


def draw_tree(node,
              child_iter=lambda n: n.children,
              text_str=str):
    return _draw_tree(node, '', child_iter, text_str)


def _draw_tree(node, prefix, child_iter, text_str):
    buf = StringIO()

    children = list(child_iter(node))

    # check if root node
    if prefix:
        buf.write(prefix[:-3])
        buf.write('  +--')
    buf.write(text_str(node))
    buf.write('\n')

    for index, child in enumerate(children):

        if index + 1 == len(children):
            sub_prefix = prefix + '   '
        else:
            sub_prefix = prefix + '  |'

        buf.write(
            _draw_tree(child, sub_prefix, child_iter, text_str)
        )

    return buf.getvalue()


def breadth_first_search(graph):
    targets = [graph.root]
    seen = {graph.root}

    while targets:
        visiting = targets.pop(0)
        print 'visiting %s' % visiting
        for child in visiting.children:
            if child not in seen:
                targets.insert(0, child)
                print 'havent seen %s, child of %s' % (child, visiting)
            graph.add_directed_edge(visiting, child)
        seen.add(visiting)
        print 'marking %s as seen' % visiting
    return seen


def convert_to_snake_case(text):
    return re.sub("(?<=[a-z])([A-Z])", "_\\1", text).lower()



import ast
import codecs
import os
import re

import setuptools
from setuptools.command.test import test as TestCommand

here = os.path.abspath(os.path.dirname(__file__))


class VersionFinder(ast.NodeVisitor):
    def __init__(self):
        self.version = None

    def visit_Assign(self, node):
        if node.targets[0].id == '__version__':
            self.version = node.value.s


def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*parts):
    finder = VersionFinder()
    finder.visit(ast.parse(read(*parts)))
    return finder.version


version = find_version('pyramid_maze', '__init__.py')

install_requires = [
    'pyramid >=1.5.1,<1.6.0',
]

extras_require = {
    'tests': [
        'pytest >=2.5.2,<2.6',
        'pytest-cov >=1.7,<2.0',
        'webtest',
        'sqlalchemy==0.9.7',
        'tox',
        'mock >=1.0,<2.0',
    ],
}

scripts = []

data_files = []


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        pytest.main(self.test_args)


packages = setuptools.find_packages('.', exclude=('tests', 'tests.*'))
setuptools.setup(
    name='pyramid_maze',
    version=version,
    url='https://github.com/mahmoudimus/pyramid_maze',
    author='mahmoudimus',
    author_email='mabdelkader@gmail.com',
    description=("A url route generator for Pyramid's traversal, based on "
                 "graph theory"),
    long_description='',
    platforms='any',
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    tests_require=extras_require['tests'],
    packages=packages,
    scripts=[],
    data_files=[],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ],
    cmdclass={'test': PyTest},
)
